/**********************************************************************
 * SimMips: Simple Computer Simulator of MIPS                         *
 * Copyright (C) 2008-2010 Arch Lab. Tokyo Institute of Technology    *
 *  Dual licensed under New BSD or GPL v2 licenses.                   *
 *  See COPYING file on the base directory for more information.      *
 **********************************************************************/
#include "define.h"

enum {
    MIERU_CNT = 0x10c,
    MIERUEMB_SW = 0x10000,
    DISPLAY_CMD_BUF_SIZE = 256,
};

/**********************************************************************/
BoardME::BoardME()
{
    lcd_path = NULL;
    lcd_size = 1;
}

/**********************************************************************/
void BoardME::usage()
{
    Board::usage();
    char usagemessage[] = "\
 [MieruEMB Mode Option]\n\
  -l [name]: specify the LCD Emulator Pipe\n\
  -2       : enlarge the image twice\n\
\n";
    printf("%s", usagemessage);
}

/**********************************************************************/
int BoardME::checkopt(char **argv, int *index)
{
    char *opt = argv[*index];
    if (opt[1] == 'l') {
        if (lcd_path) {
            fprintf(stderr, "## multiple -l options\n");
            return 1;
        }
        if ((lcd_path = argv[++*index]) == NULL) {
            fprintf(stderr, "## -l option: no file specified\n");
            return 1;
        }
        return 0;
    } else if (opt[1] == '2') {
        lcd_size = 2;
        return 0;
    }
    return Board::checkopt(argv, index);
}

/**********************************************************************/
MMDevice *BoardME::setmmdevice(char *devname, uint032_t size)
{
    if (strcmp(devname, "MIERUEMB_IO") == 0)
        return new MieruIOME();
    if (strcmp(devname, "MINILCD_VRAM") == 0)
        return new DisplayControllerME();
    return Board::setmmdevice(devname, size);
}

/**********************************************************************/
void BoardME::versioninfo()
{
    Board::versioninfo();
    printf("## - %s %s\n", L_NAME_ME, L_VER_ME);    
}

/**********************************************************************/
void MieruIOME::init(Board *b)
{
    this->board = b;
}

/**********************************************************************/
int MieruIOME::read(const uint032_t addr, uint size, void *data)
{
    // MieruPC Timer
    if (size == 4 && addr == MIERU_CNT)
        *(uint032_t *)data = (uint032_t) (board->gettime() / 1000);
    // MieruEMB Switch (always OFF)
    if (size == 4 && addr >= MIERUEMB_SW && addr <= MIERUEMB_SW + 8)
        *(uint032_t *)data = 0x01;
    if (size == 1 && addr >= MIERUEMB_SW && addr <= MIERUEMB_SW + 8 &&
        addr % 4 == 0)
        *(uint008_t *)data = 0x01;
    return 0;
}

/**********************************************************************/
DisplayControllerME::~DisplayControllerME()
{
    DELETE_ARRAY(cmdbuf);
    if (gui_pipe != -1)
        close(gui_pipe);
}

/**********************************************************************/
void DisplayControllerME::init(Board *b)
{
    BoardME *board = static_cast<BoardME *>(b);
    gui_pipe = -1;
    if (board->lcd_path) {
        int res = open(board->lcd_path, O_WRONLY);
        if (res < 0) {
            fprintf(stderr, "## Can't open LCD pipe file: %s\n",
                    board->lcd_path);
        } else {
            gui_pipe = res;
            int buff = ::write(gui_pipe, "I", 1); // LCD Reset
	    buff = buff;
        }
    }
    cmdbuf = new char[DISPLAY_CMD_BUF_SIZE + 1];
    cmdbuf[0] = '\0';

    lcd_size = board->lcd_size;
}

/**********************************************************************/
int DisplayControllerME::write(uint032_t addr, uint size, uint064_t data)
{
    int len;
    if (addr + size > 128 * 128 * 4 || size > 4)
        return 1;
    if (addr % 4)
        return 0;

    if (gui_pipe != -1) {
        len = snprintf(cmdbuf, DISPLAY_CMD_BUF_SIZE,
                       "x%03dy%03dF%02xd%s",
                       ((addr / 4) % 128) * lcd_size,
                       ((addr / 4) / 128) * lcd_size,
                       ((data & 4) ? 0xe0 : 0) |
                       ((data & 2) ? 0x1c : 0) |
                       ((data & 1) ? 0x03 : 0),
                       (lcd_size == 1) ? "" : "djhhdd");
        int buff = ::write(gui_pipe, cmdbuf, len);
	buff = buff;
    }
    return 0;
}

/**********************************************************************/
